/*
    HD44780-compatible LCD library for Atmel AVR microcontrollers
    (c) 2016 F. Štefanec
    License: Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported
    License link: https://creativecommons.org/licenses/by-nc-sa/3.0
    This file is provided without any warranty, use at your own risk.

    Include the following line in your program, if you want to use 8 bit mode:
    #define LCD_MODE_8 1
    else, the LCD will be used in 4 bit mode.
*/

#include "lcd.h"

const uint8_t lineaddrs[] = \
{
    0x00, // 16x2 display
    0x40, // line 2

    0x00, // 16x4 display
    0x40, // line 2
    0x10, // line 3
    0x50, // line 4

    0x00, // 20x4 display
    0x40, // line 2
    0x14, // line 3
    0x54  // line 4
};

void pulse_e() // pulse enable to make the display execute/
{
    LCD_CPORT |= (1<<LCD_E);
    _delay_us(1);
    LCD_CPORT &= ~(1<<LCD_E);
    _delay_us(1);
}

#ifdef LCD_MODE_8
void lcd_send(uint8_t command){ // send a command or data, 8 bit mode

    //LCD_DPORT = command;

    if(command & 0b00000001)
        LCD_DPORT |= (1<<LCD_D0);
    else
        LCD_DPORT &= ~(1<<LCD_D0);

    if(command & 0b00000010)
        LCD_DPORT |= (1<<LCD_D1);
    else
        LCD_DPORT &= ~(1<<LCD_D1);

    if(command & 0b00000100)
        LCD_DPORT |= (1<<LCD_D2);
    else
        LCD_DPORT &= ~(1<<LCD_D2);

    if(command & 0b00001000)
        LCD_DPORT |= (1<<LCD_D3);
    else
        LCD_DPORT &= ~(1<<LCD_D3);

    if(command & 0b00010000)
        LCD_DPORT |= (1<<LCD_D4);
    else
        LCD_DPORT &= ~(1<<LCD_D4);

    if(command & 0b00100000)
        LCD_DPORT |= (1<<LCD_D5);
    else
        LCD_DPORT &= ~(1<<LCD_D5);

    if(command & 0b01000000)
        LCD_DPORT |= (1<<LCD_D6);
    else
        LCD_DPORT &= ~(1<<LCD_D6);

    if(command & 0b10000000)
        LCD_DPORT |= (1<<LCD_D7);
    else
        LCD_DPORT &= ~(1<<LCD_D7);

    pulse_e();
    _delay_us(50);
}
#else
void lcd_send(uint8_t command) // send an 8-bit commands over 4-bit interface
{

    // top 4b nibble
    if(command & 0b00010000)
        LCD_DPORT |= (1<<LCD_D4);
    else
        LCD_DPORT &= ~(1<<LCD_D4);

    if(command & 0b00100000)
        LCD_DPORT |= (1<<LCD_D5);
    else
        LCD_DPORT &= ~(1<<LCD_D5);

    if(command & 0b01000000)
        LCD_DPORT |= (1<<LCD_D6);
    else
        LCD_DPORT &= ~(1<<LCD_D6);

    if(command & 0b10000000)
        LCD_DPORT |= (1<<LCD_D7);
    else
        LCD_DPORT &= ~(1<<LCD_D7);

    pulse_e();
    _delay_us(50);

    // bottom 4b nibble
    if(command & 0b00000001)
        LCD_DPORT |= (1<<LCD_D4);
    else
        LCD_DPORT &= ~(1<<LCD_D4);

    if(command & 0b00000010)
        LCD_DPORT |= (1<<LCD_D5);
    else
        LCD_DPORT &= ~(1<<LCD_D5);

    if(command & 0b00000100)
        LCD_DPORT |= (1<<LCD_D6);
    else
        LCD_DPORT &= ~(1<<LCD_D6);

    if(command & 0b00001000)
        LCD_DPORT |= (1<<LCD_D7);
    else
        LCD_DPORT &= ~(1<<LCD_D7);

    pulse_e();
    _delay_us(50);
}
#endif

void lcd_send4h(uint8_t command) // send top 4 bits (4 bit-mode)
{
    // top 4b nibble
    if(command & 0b00010000)
        LCD_DPORT |= (1<<LCD_D4);
    else
        LCD_DPORT &= ~(1<<LCD_D4);

    if(command & 0b00100000)
        LCD_DPORT |= (1<<LCD_D5);
    else
        LCD_DPORT &= ~(1<<LCD_D5);

    if(command & 0b01000000)
        LCD_DPORT |= (1<<LCD_D6);
    else
        LCD_DPORT &= ~(1<<LCD_D6);

    if(command & 0b10000000)
        LCD_DPORT |= (1<<LCD_D7);
    else
        LCD_DPORT &= ~(1<<LCD_D7);

    pulse_e();
    _delay_us(50);
}

#ifdef LCD_MODE_8
void lcd_init()
{
    lcd_send8(LCD_INITSEQ);
    _delay_ms(5);
    lcd_send8(LCD_INITSEQ);
    _delay_ms(1);
    lcd_send8(LCD_INITSEQ);
    _delay_us(50);
}
#else
void lcd_init()
{
    lcd_send4h(LCD_INITSEQ);
    _delay_ms(5);
    lcd_send4h(LCD_INITSEQ);
    _delay_ms(1);
    lcd_send4h(LCD_INITSEQ);
    _delay_us(250);
    lcd_send4h(LCD_FUNCSET);
    _delay_us(50);
}
#endif

void lcd_rshigh() // Set RS pin high - write data
{
    LCD_CPORT |= (1<<LCD_RS);
    _delay_us(1);
}

void lcd_rslow() // Set RS pin low - write commands
{
    LCD_CPORT &= ~(1<<LCD_RS);
    _delay_us(1);
}

void lcd_sendstring(char (*in_string), uint8_t len) // print a string
{
    lcd_rshigh();
    _delay_us(4);
    uint8_t i;
    for (i=0;i<len;i++)
    {
        lcd_send(in_string[i]);
    }
    lcd_rslow();
}

void lcd_setcursor(uint8_t column, uint8_t line)
{
    lcd_rslow();
    lcd_send(LCD_SETDDADD + lineaddrs[line+DISPSIZE] + column);
}

void lcd_customchar(uint8_t addr, char *customchar)
{
    uint8_t bitctr;
    for (bitctr=0;bitctr<8;bitctr++)
    {
        lcd_rslow();
        lcd_send(LCD_SETCGADD + addr*8 + bitctr);
        lcd_rshigh();
        lcd_send(customchar[bitctr]);
    }
    lcd_rslow();
}

void lcd_clear()
{
    lcd_writecmd();
    lcd_send(LCD_CLEAR);
    _delay_ms(5);
}

void lcd_printspaces(uint8_t len) // print a string
{
    lcd_writedata();
    uint8_t i;
    for (i=0;i<len;i++)
    {
        lcd_send(' ');
    }
}
